//+------------------------------------------------------------------+
//|                           Generic Optimization Formulation (GOF) |
//| Concept and Coding by Better.Trader Every.Day   Feb-2024         |
//+------------------------------------------------------------------+
#property copyright "none"
#property version   "1.00"

/*

In order to use this file in your EA you must do the following:
- Edit the EA .mq5 file
- Insert the following two lines after the input variables and before OnInit()

     ulong gof_magic= an_ulong_value;
     #include <YOUR_INCLUDE_LOCATION\GenericOptimizationFormulation.mqh>

- Save and compile your EA file

If you get compiler's errors, make sure you did:
- replace an_ulong_value by the variable containing the magic number, or by the magic numeric value
- replace YOUR_INCLUDE_LOCATION by the folder name where your include files are

*/


// include files needed for statistics and algebra
#include <Math\Stat\Lognormal.mqh>
#include <Math\Stat\Uniform.mqh>
#include <Math\Alglib\alglib.mqh>

enum gof_FunctionDefs {        // functions to build objective
   MAX_NONE=0,                 // 0] None
   MAX_AnnRetPct,              // 1] Annual Return %
   MAX_Balance,                // 2] Balance
   MAX_NetProfit,              // 3] Net Profit
   MAX_SharpeRatio,            // 4] Sharpe Ratio
   MAX_ExpPayOff,              // 5] Expected Payoff
   MAX_RecovFact,              // 6] Recovery Factor
   MAX_ProfFact,               // 7] Profit Factor
   MAX_LRcrr3,                 // 8] LRcrr^3
   MAX_NbrTradesPerWeek,       // 9] #Trades/week
   MAX_WinRatePct,             // 10] Win Rate %
   MAX_Rew2RiskRatio,          // 11] Reward/Risk(RRR=AvgWin/AvgLoss)
   MAX_OneOverLRstd,           // 12] 1/(LR std%)
   MAX_OneHoverWorstTradePct,  // 13] 100/(1+|WorstLoss/Init.Dep*100|)
   MAX_LR,                     // 14] LRslope*LRcorr/LRstd
   MAX_OneHoverEqtyMaxDDpct,   // 15] 100/(1+EqtyMaxDD%))
   MAX_StratEfficiency,        // 16] Seff=Profit/(TotalTrades*AvgLot)
   MAX_KellyCrit,              // 17] Kelly Criterion
   MAX_OneOverRoRApct          // 18] 1/Max(0.01,RoRA %)
};

enum gof_HardConstrains {
   hc_NONE=0,                     // 0] None
   hc_MaxAccountLoss_pct,         // 1] Account Loss % InitDep
   hc_maxAllowed_DDpct,           // 2] Equity DrawDown %
   hc_maxAllowednbrConLossTrades, // 3] Consecutive losing trades
   hc_minAllowedWin_pct,          // 4] Win Rate %
   hc_minAllowedNbrTradesPerWeek, // 5] # trades/week
   hc_minAllowedRecovFactor,      // 6] Recov Factor
   hc_minAllowedRRRFactor,        // 7] Reward/Risk ratio
   hc_minAllowedAnnualReturn_pct, // 8] Annual Return in %
   hc_minAllowedProfFactor,       // 9] Profit Factor
   hc_minAllowedSharpeFactor,     // 10] Sharpe Factor
   hc_minAllowedExpPayOff,        // 11] Expected PayOff
   hc_minAllowedMarginLevel,      // 12] Smallest Margin Level
   hc_maxAllowedTradeLoss,        // 13] Max Loss trade
   hc_maxAllowedRoRApct           // 14] Risk of Ruin(%)
};

enum gof_HardConstType {
   hc_GT=0, // >=     Greater or equal to
   hc_LT    // <=     Less or equal to
};



enum gof_RoRaCapital {
   roraCustomPct=0,  // Custom % of Ini.Dep.
   roraCustomAmount, // Custom Capital amount
   roraIniDep        // Initial deposit
};

enum gof_objFuncDecimals {
   fr_winRate=0, // WinRate %
   fr_MCRoRA,    // MonteCarlo Sim Risk of Ruin Account %
   fr_LRcorr,    // LR correlation
   fr_ConLoss,   // Max # Consecutive losing Trades
   fr_NONE       // None
};

input group   "=========== Custom Max Optimization =========="
input group   "- Build Custom Objective to Maximize:"

sinput gof_FunctionDefs gof_Func1   = MAX_AnnRetPct;          // Select Objective Function to Maximize 1:
sinput double gof_Target1           = 200;                    // Target 1
sinput double gof_Weight1           = 1;                      // Weight 1

sinput gof_FunctionDefs gof_Func2   = MAX_RecovFact;          // Select Objective Function to Maximize 2:
sinput double gof_Target2           = 10;                     // Target 2
sinput double gof_Weight2           = 0.1;                    // Weight 2

sinput gof_FunctionDefs gof_Func3   = MAX_NbrTradesPerWeek;   // Select Objective Function to Maximize 3
sinput double gof_Target3           = 5;                     // Target 3
sinput double gof_Weight3           = 0.1;                    // Weight 3

sinput gof_FunctionDefs gof_Func4   = MAX_WinRatePct;         // Select Objective Function to Maximize 4:
sinput double gof_Target4           = 85;                     // Target 4
sinput double gof_Weight4           = 0.25;                      // Weight 4

sinput gof_FunctionDefs gof_Func5   = MAX_Rew2RiskRatio;      // Select Objective Function to Maximize 5:
sinput double gof_Target5           = 1;                      // Target 5
sinput double gof_Weight5           = 0.01;                    // Weight 5

input group   "- Hard Constraints:"
sinput bool   gof_IncludeHardConstraints     = true;//if false, all constraints are ignored

sinput gof_HardConstrains gof_HC_1=hc_minAllowedAnnualReturn_pct; // Select Constraint Function 1:
sinput gof_HardConstType gof_HCType_1=hc_GT; // Type 1
sinput double gof_HCBound_1=50; // Bound Value 1

sinput gof_HardConstrains gof_HC_2=hc_MaxAccountLoss_pct; // Select Constraint Function 2:
sinput gof_HardConstType gof_HCType_2=hc_LT; // Type 2
sinput double gof_HCBound_2=10; // Bound Value 2

sinput gof_HardConstrains gof_HC_3=hc_maxAllowed_DDpct; // Select Constraint Function 3:
sinput gof_HardConstType gof_HCType_3=hc_LT; // Type 3
sinput double gof_HCBound_3=10; // Bound Value 3

sinput gof_HardConstrains gof_HC_4=hc_maxAllowednbrConLossTrades; // Select Constraint Function 4:
sinput gof_HardConstType gof_HCType_4=hc_LT; // Type 4
sinput double gof_HCBound_4=5; // Bound Value 4

sinput gof_HardConstrains gof_HC_5=hc_minAllowedWin_pct; // Select Constraint Function 5:
sinput gof_HardConstType gof_HCType_5=hc_GT; // Type 5
sinput double gof_HCBound_5=50; // Bound Value 5

sinput gof_HardConstrains gof_HC_6=hc_minAllowedNbrTradesPerWeek; // Select Constraint Function 6:
sinput gof_HardConstType gof_HCType_6=hc_GT; // Type 6
sinput double gof_HCBound_6=2; // Bound Value 6

sinput gof_HardConstrains gof_HC_7=hc_maxAllowedRoRApct; // Select Constraint Function 7:
sinput gof_HardConstType gof_HCType_7=hc_LT; // Type 7
sinput double gof_HCBound_7=1; // Bound Value 7

sinput gof_HardConstrains gof_HC_8=hc_NONE; // Select Constraint Function 8:
sinput gof_HardConstType gof_HCType_8=hc_LT; // Type 8
sinput double gof_HCBound_8=0; // Bound Value 8

sinput gof_HardConstrains gof_HC_9=hc_NONE; // Select Constraint Function 9:
sinput gof_HardConstType gof_HCType_9=hc_LT; // Type 9
sinput double gof_HCBound_9=0; // Bound Value 9

sinput gof_HardConstrains gof_HC_10=hc_NONE; // Select Constraint Function 10:
sinput gof_HardConstType gof_HCType_10=hc_LT; // Type 10
sinput double gof_HCBound_10=0; // Bound Value 10

input group   "------ Misc Optimization Params -----"
sinput gof_objFuncDecimals gof_fr                  = fr_winRate;     // Choose Result-column's decimals
sinput gof_RoRaCapital  gof_roraMaxCap             = roraCustomPct;  // Choose capital method for Risk of Ruin
sinput double           gof_RoraCustomValue        = 10;             // Custom Value for Risk of Ruin (if needed)
sinput bool             gof_drawSummary            = false;          // Draw summary on chart
sinput bool             gof_printSummary           = true;           // Print summary on journal
sinput bool             gof_discardLargestProfit   = false;          // Subtract Largest Profit from Netprofit
sinput bool             gof_discardLargestLoss     = false;          // Add Largest Loss to Net profit
sinput double           gof_PenaltyMultiplier      = 100;            // Multiplier for Penalties (k_p)
sinput double           gof_ObjMultiplier          = 100;            // Multiplier for Objectives (k_o)

enum ENUM_CORR_TYPE {
   CORR_PEARSON,     // Pearson's correlation
   CORR_SPEARMAN     // Spearman's Rank-Order correlation
};

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double       gof_ObjPositiveScalefactor=1e6;
double       gof_ObjNegativeScalefactor=1e3;
double       gof_PropFirmInitialDeposit;
datetime     gof_finalDate;
long         gof_totalTime;
double       gof_nyears;
int          gof_profitTrades;
int          gof_lossTrades;
double       gof_totalProfit;
double       gof_totalLoss;
double       gof_initialDeposit;
int          gof_nbrContProfitTrades;
int          gof_nbrAvgConLossTrades;
int          gof_nbrConLossTrades;
double       gof_contLossMax;
double       gof_minBalance;
double       gof_minEquity;
double       gof_maxBalDDrelPct;
double       gof_maxBalDD;
double       gof_maxBalDDrel;
double       gof_maxEquiDDrel;
double       gof_maxEquiDD;
double       gof_maxEquiDDrelPct;
double       gof_posDealsStd;
double       gof_negDealsStd;
double       gof_profitFactor;
double       gof_recovFactor;
double       gof_ExpectedPayOff;
double       gof_maxLossTrade;
double       gof_maxProfitTrade;
double       gof_endBalance;
double       gof_AnnualReturnPct;
int          gof_totTrades;
double       gof_netProfit;
double       gof_avgVol;
double       gof_largestVol;
double       gof_netProfitTS;
double       gof_winPct;
double       gof_sharpeRatio;
double       gof_RewardRiskRatio;
double       gof_largestProfitTrade=0;

double       gof_Targets[]= {gof_Target1,gof_Target2,gof_Target3,gof_Target4,gof_Target5};
double       gof_Weights[]= {gof_Weight1,gof_Weight2,gof_Weight3,gof_Weight4,gof_Weight5};
gof_FunctionDefs gof_Functions[]= {gof_Func1,gof_Func2,gof_Func3,gof_Func4,gof_Func5};
double       gof_FunctionValue[]= {0.,0.,0.,0.,0.};
double       gof_Contrib2SingleObjective[]= {0.,0.,0.,0.,0.};
gof_HardConstrains gof_HCFunctions[]= {gof_HC_1,gof_HC_2,gof_HC_3,gof_HC_4,gof_HC_5,gof_HC_6,gof_HC_7,gof_HC_8,gof_HC_9,gof_HC_10};
gof_HardConstType gof_HCType[]= {gof_HCType_1,gof_HCType_2,gof_HCType_3,gof_HCType_4,gof_HCType_5,gof_HCType_6,gof_HCType_7,gof_HCType_8,gof_HCType_9,gof_HCType_10};
double gof_HCBounds[]= {gof_HCBound_1,gof_HCBound_2,gof_HCBound_3,gof_HCBound_4,gof_HCBound_5,gof_HCBound_6,gof_HCBound_7,gof_HCBound_8,gof_HCBound_9,gof_HCBound_10};

int          gof_labelX              = 20;
int          gof_labelY              = 25;
const int    gof_labelHeight         = 15;
string       gof_labelText           = "";
string       gof_summaryLine[50];
int          gof_summaryLineCounter=0;
double       gof_avgProfit=0;
double       gof_avgLoss=0;
double       gof_LRstd=0;
double       gof_minMarginLevel;
double       gof_maxEquiDDPct;
double       gof_maxBalDDPct;
double       gof_LRcorr;
double       gof_LRintercept;
double       gof_LRslope;
string       gof_symbolName= Symbol();
bool         gof_expertRemoved;
double       gof_totalCommission;
double       gof_totalVolume;
double       gof_RoRApct;
double       gof_dealsEquity[];
bool         gof_ExpertRemoved=false;
ENUM_TIMEFRAMES gof_timeFrame=Period();
datetime     gof_startingDate=iTime(gof_symbolName,gof_timeFrame,0);// simulation starting date
double       gof_minReqMarginLevel;
double       gof_RoundTripCommisssionPerLot=4;// in account currency (change this as needed)
double       gof_constraintTotalPenalty;
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double OnTester() {

// collecting Tester Statistics
   gof_netProfitTS         = TesterStatistics(STAT_PROFIT);
   gof_totTrades           = (int)TesterStatistics(STAT_TRADES);
   if(gof_totTrades<=1)return(EMPTY_VALUE);//-MathAbs(gof_ObjNegativeScalefactor*3500000));
   gof_minReqMarginLevel   = TesterStatistics(STAT_MIN_MARGINLEVEL);
   gof_finalDate           = iTime(gof_symbolName,gof_timeFrame,0);
   gof_totalTime           = (long)gof_finalDate-(long)gof_startingDate;
   gof_nyears              = (double)gof_totalTime/3600./24./365.;
   gof_profitTrades        = (int) TesterStatistics(STAT_PROFIT_TRADES);
   gof_lossTrades          = (int) TesterStatistics(STAT_LOSS_TRADES);
   gof_totalProfit         = TesterStatistics(STAT_GROSS_PROFIT);
   gof_totalLoss           = TesterStatistics(STAT_GROSS_LOSS);
   gof_largestProfitTrade  = TesterStatistics(STAT_MAX_PROFITTRADE);
   gof_initialDeposit      = TesterStatistics(STAT_INITIAL_DEPOSIT);
   gof_nbrContProfitTrades = (int)TesterStatistics(STAT_CONPROFITMAX_TRADES);
   gof_nbrAvgConLossTrades = (int)TesterStatistics(STAT_LOSSTRADES_AVGCON);
   gof_nbrConLossTrades    = (int)TesterStatistics(STAT_MAX_CONLOSS_TRADES);
   gof_contLossMax         = TesterStatistics(STAT_CONLOSSMAX);
   gof_minBalance          = TesterStatistics(STAT_BALANCEMIN);
   gof_minEquity           = TesterStatistics(STAT_EQUITYMIN);
   gof_maxBalDDrelPct      = TesterStatistics(STAT_BALANCE_DDREL_PERCENT);
   gof_maxBalDD            = TesterStatistics(STAT_BALANCE_DD);
   gof_maxBalDDPct         = TesterStatistics(STAT_BALANCEDD_PERCENT);
   gof_maxBalDDrel         = TesterStatistics(STAT_BALANCE_DD_RELATIVE);
   gof_maxEquiDDrel        = TesterStatistics(STAT_EQUITY_DD_RELATIVE);
   gof_maxEquiDD           = TesterStatistics(STAT_EQUITY_DD);
   gof_maxEquiDDPct        = TesterStatistics(STAT_EQUITYDD_PERCENT);
   gof_maxEquiDDrelPct     = TesterStatistics(STAT_EQUITY_DDREL_PERCENT);
   gof_profitFactor        = TesterStatistics(STAT_PROFIT_FACTOR);
   gof_recovFactor         = TesterStatistics(STAT_RECOVERY_FACTOR);
   gof_ExpectedPayOff      = TesterStatistics(STAT_EXPECTED_PAYOFF);
   gof_maxLossTrade        = TesterStatistics(STAT_MAX_LOSSTRADE);
   gof_maxProfitTrade      = TesterStatistics(STAT_MAX_PROFITTRADE);
   gof_minMarginLevel      = TesterStatistics(STAT_MIN_MARGINLEVEL);
   gof_sharpeRatio         = TesterStatistics(STAT_SHARPE_RATIO);

// additional statistics
   gof_netProfit           = calcNetProfitAfterExpenses(gof_avgVol,gof_largestVol,gof_totalCommission,gof_totalVolume);
   gof_endBalance          = gof_initialDeposit+gof_netProfit;//gof_totalProfit+gof_totalLoss;
   gof_AnnualReturnPct     = gof_nyears==0?0:(MathPow(gof_endBalance/gof_initialDeposit,1./gof_nyears)-1)*100.;
   gof_AnnualReturnPct     = MathIsValidNumber(gof_AnnualReturnPct)?gof_AnnualReturnPct:-100;
   gof_winPct              = (gof_totTrades!=0)?100.*gof_profitTrades/gof_totTrades:0.;

   gof_RewardRiskRatio     = gof_profitTrades*gof_totalLoss!=0?MathAbs((gof_totalProfit*gof_lossTrades)/(gof_profitTrades*gof_totalLoss)):0;
   gof_RewardRiskRatio     = gof_winPct>=100?10:gof_RewardRiskRatio;

   gof_avgProfit           = gof_profitTrades>0?gof_totalProfit/gof_profitTrades:0;
   gof_avgLoss             = gof_lossTrades>0?gof_totalLoss/gof_lossTrades:0;
   calcDealsEquity(gof_dealsEquity); // get the equity history
   gof_LRcorr              = CustomR2Equity(gof_dealsEquity,gof_LRstd,gof_LRintercept,gof_LRslope,CORR_PEARSON);
   bool gof_displayContraintFlag    = true;
   int  CHART_ID                    = 0;

// risk of ruin calculation
   double roracap=0;
   if(gof_roraMaxCap==roraIniDep)         roracap=gof_initialDeposit;
   if(gof_roraMaxCap==roraCustomAmount)   roracap=gof_RoraCustomValue; // in account currency
   if(gof_roraMaxCap==roraCustomPct)      roracap=gof_RoraCustomValue*gof_initialDeposit/100;
   gof_RoRApct = gof_winPct>0?
                 MonteCarlo_RiskOfRuinAccount(gof_winPct,gof_avgProfit,gof_avgLoss,roracap,gof_totTrades):
                 99;
  
//------------ GOF ----------------------
// Printing and displaying results from the simulation
   GOFsummaryReport();

// calculate the single objective function
   double SingleObjective = calcObjFunc();


// calculate the total penalty from constraint violations
   if(gof_IncludeHardConstraints) gof_constraintTotalPenalty=calcContraintTotalPenalty(gof_displayContraintFlag);

// Compute customMaxCriterion
// gof_PenaltyMultiplier pushes infeasible designs to have low objective values
// gof_PenaltyMultiplier expand the positive side of the Y axis
   double customMaxCriterion=gof_constraintTotalPenalty>0?
                             SingleObjective-gof_PenaltyMultiplier*gof_constraintTotalPenalty:
                             gof_ObjMultiplier*SingleObjective;

// add additional simulation result as two decimal digits in the result column
   customMaxCriterion=AddDecimalsToCustomMax(customMaxCriterion);

// Printing and displaying more results from GOF
   FinishGOFsummaryReport(customMaxCriterion);

   return (NormalizeDouble(customMaxCriterion,2));
}
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double calcObjFunc() {

   double SumOfWeights=0;
   double obj2max=0.;
   int n=0;
   for(int i=0; i<=ArraySize(gof_FunctionValue)-1; i++)
      n=gof_Functions[i]!=MAX_NONE?n+1:n;

   if(n==0) {
      Print("**** ERROR in OnTester: no functions selected for the Optimizer");
      return(0);
   }

   double sumOfObjFunc=0;
   double iWeight,iTarget,r;
   for(int i=0; i<=ArraySize(gof_FunctionValue)-1; i++) {
      iWeight=gof_Weights[i];
      iTarget=gof_Targets[i];
      switch(gof_Functions[i]) {

      case MAX_NONE:
         obj2max=0;
         iWeight=0;
         iTarget=1;
         break;

      case MAX_AnnRetPct:
         obj2max=gof_AnnualReturnPct;
         break;

      case MAX_Balance:
         obj2max=gof_endBalance;
         break;

      case MAX_WinRatePct:
         obj2max=gof_winPct;
         break;

      case MAX_NbrTradesPerWeek:
         obj2max=gof_totTrades/(gof_nyears*52);
         break;

      case MAX_NetProfit:
         obj2max=gof_netProfit;
         break;

      case MAX_LR:
         obj2max=gof_LRslope*gof_LRcorr/gof_LRstd;
         break;

      case MAX_SharpeRatio:
         obj2max=gof_sharpeRatio;
         break;

      case MAX_ExpPayOff:
         obj2max=gof_ExpectedPayOff;
         break;

      case MAX_RecovFact:
         obj2max=gof_recovFactor;
         break;

      case MAX_ProfFact:
         obj2max=gof_profitFactor;
         break;

      case MAX_LRcrr3:
         obj2max=MathPow(gof_LRcorr,3);
         break;

      case MAX_OneHoverWorstTradePct:
         obj2max=100./(1+MathAbs(gof_maxLossTrade/gof_initialDeposit*100));
         break;

      case MAX_Rew2RiskRatio:
         obj2max=gof_RewardRiskRatio;
         break;

      case MAX_OneOverLRstd:
         obj2max=1/gof_LRstd;
         break;

      case MAX_OneHoverEqtyMaxDDpct:
         obj2max=100./(1.+MathAbs(gof_maxEquiDDPct));
         break;

      case MAX_StratEfficiency:
         obj2max=gof_netProfit/gof_totTrades/gof_avgVol;
         break;

      case MAX_KellyCrit:
         r=gof_lossTrades!=0?gof_profitTrades/gof_lossTrades:0; // Win/loss ratio;
         obj2max=r!=0?gof_winPct-((100-gof_winPct)/r):0; // Kelly Criterion
         break;

      case MAX_OneOverRoRApct:
         obj2max=1/MathMax(0.01,gof_RoRApct); // Risk of Ruin Criterion
         break;
      }

      gof_FunctionValue[i]=obj2max;
      SumOfWeights+=iWeight;
      gof_Contrib2SingleObjective[i]=iWeight/iTarget*obj2max;

      sumOfObjFunc+=gof_Contrib2SingleObjective[i];
   }


// add obj results to the summary report

   gof_summaryLine[gof_summaryLineCounter++]="Objective Functions:";
   gof_summaryLine[gof_summaryLineCounter++]="Value Target Weight Contribution%";

   double sumOfContribution=0;
   for(int i=0; i<=ArraySize(gof_FunctionValue)-1; i++)
      if(gof_Functions[i]!=MAX_NONE)
         sumOfContribution+=gof_Contrib2SingleObjective[i];

   for(int i=0; i<=ArraySize(gof_FunctionValue)-1; i++) {
      if(gof_Functions[i]!=MAX_NONE) {
         gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.2f %.2f %.2f %.1f%%| %s",
               gof_FunctionValue[i], gof_Targets[i],gof_Weights[i],
               gof_Contrib2SingleObjective[i]/sumOfContribution*100,EnumToString(gof_Functions[i]));;
      }
   }
   gof_summaryLine[gof_summaryLineCounter++]=" ";
   gof_summaryLine[gof_summaryLineCounter++]="Constraints: (Actual vs Bound)";
   if(!gof_IncludeHardConstraints)gof_summaryLine[gof_summaryLineCounter++]="------- None!";


// return obj with convexification of weights
   return(sumOfObjFunc/SumOfWeights);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double calcContraintTotalPenalty(bool gof_displayContraintFlag) {
   double penaltySum=0;
   for(int i=0; i<=ArraySize(gof_HCFunctions)-1; i++) {

      switch(gof_HCFunctions[i]) {

      case hc_NONE: // no constraint
         break;

      case hc_MaxAccountLoss_pct:         // Account Loss % InitDep allowed
         penaltySum+=Violation_C_MaxAccLoss(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      case hc_maxAllowed_DDpct:          // Eqty DD% allowed
         penaltySum+=Violation_C_DrawDown(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      case hc_maxAllowednbrConLossTrades: // consctiv losing trades allowed
         penaltySum+=Violation_C_ConLossTrd(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      case hc_minAllowedWin_pct:          // % wins/total trades allowed
         penaltySum+=Violation_C_WinRate(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      case hc_minAllowedNbrTradesPerWeek: // # trades/week allowed
         penaltySum+=Violation_C_NbrTrdPerWeek(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      case hc_minAllowedRecovFactor:      // Recov Factor allowed
         penaltySum+=Violation_C_RecFac(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      case hc_minAllowedRRRFactor:        // Reward/Risk allowed
         penaltySum+=Violation_C_RRR(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      case hc_minAllowedAnnualReturn_pct: // Annual Return allowed in %
         penaltySum+=Violation_C_AnnRet(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      case hc_minAllowedProfFactor:       // Profit Factor allowed
         penaltySum+=Violation_C_ProfFactor(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      case hc_minAllowedSharpeFactor:     // Sharpe Factor allowed
         penaltySum+=Violation_C_Sharpe(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      case hc_minAllowedExpPayOff:      // Expected PayOff allowed
         penaltySum+=Violation_C_ExpPayOff(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      case hc_minAllowedMarginLevel:      // Margin Level allowed
         penaltySum+=Violation_C_MarginLevel(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      case hc_maxAllowedTradeLoss:        // Loss trade allowed in $
         penaltySum+=Violation_C_MaxTradeLoss(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      case hc_maxAllowedRoRApct:           // Risk of Ruin(%) allowed
         penaltySum+=Violation_C_MaxRoRpct(gof_HCType[i],gof_HCBounds[i], gof_displayContraintFlag);
         break;

      }
   }

   return(penaltySum);
}

//------------------- HARD CONSTRAINT FUNCTIONS -------------------
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_MaxAccLoss(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   hc_bound=-MathAbs(hc_bound);// forcing the bound to be negative
   double v=(1-gof_minEquity/gof_initialDeposit)*100;
   double res=2*calcConstraintPenalty(v,hc_GT,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : MaxAccLoss = %.2f%% vs %.2f%%", res<=0?"pass  ":"FAIL !", -v,hc_bound);
   return(res);
}


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_DrawDown(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   double v=gof_maxEquiDDPct;
   double res=calcConstraintPenalty(v,hc_type,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : Eqty_DrawDown = %.2f%% vs %.2f%%",res<=0?"pass  ":"FAIL !",v,hc_bound);
   return(res);
}


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_MaxTradeLoss(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   double v=gof_maxLossTrade;
   double res=calcConstraintPenalty(v,hc_type,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : maxTradeLoss = $%.0f vs %.0f", res<=0?"pass  ":"FAIL !",gof_maxLossTrade,-MathAbs(hc_bound));
   return(res);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_MaxRoRpct(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   double v=gof_RoRApct;
   double res=calcConstraintPenalty(v,hc_type,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : RiskOfRuinAcc = %.2f%% vs %.2f%%", res<=0?"pass  ":"FAIL !",v,hc_bound);
   return(res);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_ConLossTrd(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   int v=(int)gof_nbrConLossTrades;
   double res=calcConstraintPenalty(v,hc_type,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : nbrConLossTrades = %d vs %.0f", res<=0?"pass  ":"FAIL !",v,hc_bound);
   return(res);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_WinRate(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   double v=gof_winPct;
   double res=calcConstraintPenalty(v,hc_type,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : WinRate = %.0f%% vs %.0f%%", res<=0?"pass  ":"FAIL !",v,hc_bound);
   return(res);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_RecFac(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   double v=gof_recovFactor;
   double res=calcConstraintPenalty(v,hc_type,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : RecovFactor = %.2f vs %.2f", res<=0?"pass  ":"FAIL !",v,hc_bound);
   return(res);
}


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_Sharpe(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   double v=gof_sharpeRatio;
   double res=calcConstraintPenalty(v,hc_type,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : SharpeRatio = %.2f vs %.2f", res<=0?"pass  ":"FAIL !",v,hc_bound);
   return(res);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_RRR(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   double v=gof_RewardRiskRatio;
   double res=calcConstraintPenalty(v,hc_type,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : RewardRiskRatio = %.2f vs %.2f ", res<=0?"pass  ":"FAIL !",v,hc_bound);
   return(res);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_AnnRet(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   double v=gof_AnnualReturnPct;
   double res=calcConstraintPenalty(v,hc_type,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : AnnualReturn = %.1f%% vs %.1f%% ", res<=0?"pass  ":"FAIL !",v,hc_bound);
   return(res);
}


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_NbrTrdPerWeek(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   double v=gof_totTrades/(gof_nyears*52);
   double res=calcConstraintPenalty(v,hc_type,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : Trades/Week = %.1f vs %.1f", res<=0?"pass  ":"FAIL !",v,hc_bound);
   return(res);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_ProfFactor(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   double v=gof_profitFactor;
   double res=calcConstraintPenalty(v,hc_type,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : ProfFactor = %.2f vs %.2f", res<=0?"pass  ":"FAIL !",v,hc_bound);
   return(res);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_ExpPayOff(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   double v=gof_ExpectedPayOff;
   double res=calcConstraintPenalty(v,hc_type,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : ExpPayOff = %.2f vs %.2f", res<=0?"pass  ":"FAIL !",v,hc_bound);
   return(res);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Violation_C_MarginLevel(gof_HardConstType hc_type, double hc_bound, bool displayContraint=true) {
   double v=gof_minMarginLevel;
   double res=calcConstraintPenalty(v,hc_type,hc_bound);
   if(displayContraint)
      gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%s : MarginLevel = %.0f%% vs %.0f%%", res<=0?"pass  ":"FAIL !",v,hc_bound);
   return(res);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void calcDealsEquity(double &deals_equity[]) {
// this fills the deals_equity array based on the history of trades
   HistorySelect(gof_startingDate, gof_finalDate);
   double std=0;
   double sum_profit = 0;

   int count = 0;
   int total = HistoryDealsTotal();
   for(int i = 0; i < total; i++) {
      ulong ticket = HistoryDealGetTicket(i);

      double profit = HistoryDealGetDouble(ticket, DEAL_PROFIT);

      if(profit == 0.0) continue;
      if(ArraySize(deals_equity) <= count)ArrayResize(deals_equity, count+20);
      sum_profit += profit;
      deals_equity[count] = sum_profit;
      count++;
   }
   ArrayResize(deals_equity, count);
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double calcConstraintPenalty(double g_value, gof_HardConstType hc_type, double hc_bound) {
// calculate the penalty for violating a constraint
   double penalty=hc_type==hc_LT?
                  MathMax(0,(g_value-hc_bound)/(hc_bound==0?1:MathAbs(hc_bound))):
                  MathMax(0,(hc_bound-g_value)/(hc_bound==0?1:MathAbs(hc_bound)));
   return(penalty);
}

//+------------------------------------------------------------------+
//| Returns the R^2 estimate based on the equity signal              |
//+------------------------------------------------------------------+
double CustomR2Equity(double& equity[], double& std, double &intercept, double &slope, ENUM_CORR_TYPE corr_type = CORR_PEARSON) {
// calculate R^2 for the Equity line. This function is a modification of a function found in CodeBase.


   int total = ArraySize(equity);
   if(total <= 1)
      return 0.0;

//-- Fill the matrix: Y - equity value, X - ordinal number of the value
   CMatrixDouble xy(total, 2);
   for(int i = 0; i < total; i++) {
      xy.Set(i,0,i);
      xy.Set(i,1,equity[i]);

   }
//-- Find coefficients a and b of the linear model y = a*x + b;
   int retcode = 0;

   CLinReg::LRLine(xy, total, retcode, intercept, slope);
//-- Generate the linear regression values for each X;
   double estimate[];
   ArrayResize(estimate, total);

   std=0;

   for(int x = 0; x < total; x++) {
      estimate[x]=intercept+x*slope;
      std = std+MathPow((estimate[x]-equity[x])/estimate[x],2);
   }
   if(MathAbs(std)<0.000001)return(0.);
   std=100*MathPow(std/(total-1),0.5);
   std=std<=0?TesterStatistics(STAT_INITIAL_DEPOSIT)*10:std;

//-- Find the coefficient of correlation of values with their linear regression
   double corr = 0.0;
   if(corr_type == CORR_PEARSON)
      corr = CAlglib::PearsonCorr2(equity, estimate);
   else
      corr = CAlglib::SpearmanCorr2(equity, estimate);
//-- Find R^2 and its sign
   double r2 = MathPow(corr, 2.0);
   r2=equity[0] > equity[total-1]?-r2:r2;

////-- Return the R^2 estimate normalized to 3 decimals
   return NormalizeDouble(r2,3);

}



//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double calcNetProfitAfterExpenses(double &avgVol, double &largestVol, double &TotalComm, double &totalVolume) {
// Calculate Net profit after expenses (commision, swap)
   HistorySelect(gof_startingDate, gof_finalDate);// select day's history
   TotalComm=0; // OUT deal's profits (closed deals)
   double myTotalComm=0;
   int totalDeals = HistoryDealsTotal();
   int totalOutDeals=0;
   double dealVol=0;
   ulong dealMagic;
   largestVol=-1;
   avgVol=0;
   totalVolume=0;

   for(int i = 0; i < totalDeals; i++) {
      ulong ticket = HistoryDealGetTicket(i);
      dealMagic=HistoryDealGetInteger(ticket,DEAL_MAGIC);
      ENUM_DEAL_ENTRY dealEntry = (ENUM_DEAL_ENTRY) HistoryDealGetInteger(ticket, DEAL_ENTRY);
      double deal_comm=HistoryDealGetDouble(ticket,DEAL_COMMISSION);
      dealVol=HistoryDealGetDouble(ticket,DEAL_VOLUME);

      // swap expenses is accounted in the net profit already, but commissions may not.
      if(dealEntry==DEAL_ENTRY_OUT && dealMagic==gof_magic) {
         totalOutDeals++;
         TotalComm = TotalComm+deal_comm;
         myTotalComm=myTotalComm-dealVol*gof_RoundTripCommisssionPerLot;
         largestVol=MathMax(largestVol,dealVol);
         totalVolume+=dealVol;
      }
   }

   avgVol=totalOutDeals>0?totalVolume/totalOutDeals:0;

// accounting for TotalComm
// user may want to penalize net profit with largest win or loss
   double result=TesterStatistics(STAT_PROFIT) + TotalComm
                 -gof_discardLargestProfit*gof_largestProfitTrade
                 +gof_discardLargestLoss*gof_maxLossTrade;

   return(result);

}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void CalcDealStatistics(double &dealsEquity[], double &posDealsMean,double &posDealsStd,double &negDealsMean,double &negDealsStd) {

// calculate mean and stdev of wins and losses from the trade history in dealsEquity

   int n=ArraySize(dealsEquity);
   double posDeals[];
   double negDeals[];
   ArrayResize(posDeals,n);
   ArrayResize(negDeals,n);
   int ipos=0,ineg=0;
   for(int i=0; i<n-1; i++) {
      double profit=dealsEquity[i+1]-dealsEquity[i];
      if(profit>=0) {
         posDeals[ipos]=profit;
         ipos++;
      } else {
         negDeals[ineg]=profit;
         ineg++;
      }
   }
   ArrayResize(posDeals,ipos);
   ArrayResize(negDeals,ineg);

   posDealsMean=MathMean(posDeals);
   posDealsStd=ipos>1?MathStandardDeviation(posDeals):0;
   negDealsMean=MathMean(negDeals);
   negDealsStd=ineg>1?MathStandardDeviation(negDeals):0;
   gof_posDealsStd=posDealsStd;
   gof_negDealsStd=negDealsStd;
}

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double MonteCarlo_RiskOfRuinAccount(double WinRatePct, double AvgWin, double AvgLoss, double limitLoss_money, int nTrades) {
// 10000 Montecarlo simulations, each with at least 100 trades.
// Ideally, if we had lots of trades in the history, we could use a Markov Chain transfer probability matrix
//  we are limiting the statistics to mean & stdev, without knowledge of a transfer probability information

   double posDealsMean,posDealsStd,negDealsMean,negDealsStd;
   CalcDealStatistics(gof_dealsEquity, posDealsMean,posDealsStd,negDealsMean,negDealsStd);

// seeding the random number generator
   MathSrand((int)TimeLocal()+1);

// ignore posDealsMean and negDealsMean. Use AvgWin and AvgLoss instead
   AvgLoss=MathAbs(AvgLoss);
   WinRatePct=MathMin(100,MathMax(0,WinRatePct));

// case when win rate is 100%:
   if((int)(WinRatePct*nTrades/100)>=nTrades) {
      WinRatePct=99;          // just to be a bit conservative if winrate=100%
      AvgLoss=AvgWin/2;       // a guessengineering value
      negDealsStd=posDealsStd;// a guessengineering value
   }

// Use log-normal distribution function. Mean and Std are estimated as:
   double win_lnMean =log(AvgWin*AvgWin/sqrt(AvgWin*AvgWin+posDealsStd*posDealsStd));
   double loss_lnMean=log(AvgLoss*AvgLoss/sqrt(AvgLoss*AvgLoss+negDealsStd*negDealsStd));
   double win_lnstd  =sqrt(log(1+(posDealsStd*posDealsStd)/(AvgWin*AvgWin)));
   double loss_lnstd =sqrt(log(1+(negDealsStd*negDealsStd)/(AvgLoss*AvgLoss)));

   double rand_Win[],rand_Loss[];
   double r[];

// limit amount of money that defines Ruin
   limitLoss_money=MathAbs(limitLoss_money);
   bool success;
   int ruinCount=0; // counter of ruins
   int successfulMCcounter=0;
   int nTradesPerSim=MathMax(100,nTrades);// at least 100 trades per sim
   int nMCsims=10000; // MC sims, each one with nTradesPerSim

   for(int iMC=0; iMC<nMCsims; iMC++) {
      success=MathRandomUniform(0,1,nTradesPerSim,r);

      // generate nTradesPerSim wins and losses for each simulation
      // use LogNormal distribution
      success&=MathQuantileLognormal(r,win_lnMean,win_lnstd,true,false,rand_Win);
      success&=MathQuantileLognormal(r,loss_lnMean,loss_lnstd,true,false,rand_Loss);
      if(!success)continue;
      successfulMCcounter++;
      //simulate nTradesPerSim
      double eqty=0; // start each simulation with zero equity
      for(int i=0; i<nTradesPerSim; i++) {

         // draw a random number in [0,1]
         double randNumber=(double)MathRand()/32767.;

         // select a win or a loss depending on the win rate and the random number
         // and add to the equity
         eqty+=randNumber*100 < WinRatePct?
               rand_Win[i]:
               -rand_Loss[i];

         // check if equity is below the limit (ruin)
         // count the number of times there is a ruin
         if(eqty<= -limitLoss_money) {
            ruinCount++;
            break;
         }
      }
   }
// compute risk of ruin as percentage
   double RiskOfRuinPct=(double)(ruinCount)/successfulMCcounter*100.;
   return(RiskOfRuinPct);
}

//+------------------------------------------------------------------+
double AddDecimalsToCustomMax(double obj) {
// We take advantage of having two decimal positions reported in the
// results' colum during the optimization
// We simply multiply the objective by a large power of 10, floor the
// the resulting value, and then add a quantity with
// limited range of [0,100] or [0,1]. See below.

   double dec=0;

   if(gof_fr==fr_NONE)return(NormalizeDouble(obj,2));

   switch(gof_fr) {

   case fr_winRate:
      dec=gof_winPct;
      if(gof_profitTrades==gof_totTrades)dec=99; // use 99 for 100% winrate
      break;

   case fr_MCRoRA:
      dec=gof_RoRApct;
      if(dec>=100)dec=99;// use 99 for 100% Risk of Ruin Account
      break;

   case fr_LRcorr:
      dec=gof_LRcorr*100;
      if(dec>=100)dec=99;
      if(dec<0)dec=0; // negative LRcorr will have the wrong sign
      break;

   case fr_ConLoss:
      dec=gof_nbrConLossTrades;
      if(dec>=100)dec=99; // beyond 99 consecutive losses, we use 99
      break;
   }

// positive objectives are multiplied by 1000 when they are positive
// only to have better y-scale for input variable in 1D plots
   obj=obj>0?
       MathFloor(obj*gof_ObjPositiveScalefactor)+dec/100.:
       -(MathFloor(-obj*gof_ObjNegativeScalefactor)+dec/100.);

   return(NormalizeDouble(obj,2));
}
//+------------------------------------------------------------------+
void GOFsummaryReport() {

   gof_summaryLine[gof_summaryLineCounter++]= StringFormat("%.1f %%  %.2f years | Ann.Profit TestLength",
         gof_AnnualReturnPct, gof_nyears);

   gof_summaryLine[gof_summaryLineCounter++]= StringFormat("%.0f / $%.0f | Prof(inc Expenses)/Init Dep",
         gof_netProfit,gof_initialDeposit);

   gof_summaryLine[gof_summaryLineCounter++]= StringFormat("%.1f | MetaTrade Profit",
         gof_netProfitTS);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.1f(%.1f%%), %.1f(%.1f%%) | minEqty-IniDep, minBal-IniDep",
         gof_minEquity-gof_initialDeposit, 100.*(gof_minEquity-gof_initialDeposit)/gof_initialDeposit,
         gof_minBalance-gof_initialDeposit,100.*(gof_minBalance-gof_initialDeposit)/gof_initialDeposit);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.1f %.2f%% | maxLossTrde RiskOfRuin",
         gof_maxLossTrade,gof_RoRApct);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.1f%% %.1f %.1f  | %%mxBalDDrel mxBalDDrel mxBalDD",
         gof_maxBalDDrelPct,gof_maxBalDDrel,gof_maxBalDD);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.1f%%, $%.1f $%.1f  | %%mxEqDDrel mxEqDDrel mxEqDD",
         gof_maxEquiDDrelPct,gof_maxEquiDDrel,gof_maxEquiDD);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.2f %.2f %.2f | ProfFact sharpeRtio LRstd%%",
         gof_profitFactor,gof_sharpeRatio,gof_LRstd);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.2f %.0f %.2f | LRcorr LRintercept LRslope",
         gof_LRcorr,gof_LRintercept,gof_LRslope);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.0f%% %d/%d %d | %%WinRate Wins/TotTrd #Losses",
         gof_winPct,gof_profitTrades,gof_totTrades,gof_totTrades-gof_profitTrades);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.2f %.2f $%.1f | RRR RecFct ExpPayOff",
         gof_RewardRiskRatio,gof_recovFactor,gof_ExpectedPayOff);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.2f %.2f | AvgVol LargestVol",
         gof_avgVol,gof_largestVol);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%d %d   | #AvgLossTrd  #ContLossTrd",
         gof_nbrAvgConLossTrades,gof_nbrConLossTrades);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.1f %.1f    | BestTradeGain WorstTradeLoss",
         gof_maxProfitTrade,gof_maxLossTrade);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.1f %.1f    | AvgProfit AvgLoss",
         gof_avgProfit,gof_avgLoss);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.1f %.1f    | WinStd LossStd",
         gof_posDealsStd,gof_negDealsStd);

   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("1:%d %.0f%%    | AccLeverage MinMargLevel",
         AccountInfoInteger(ACCOUNT_LEVERAGE), gof_minMarginLevel);

   gof_summaryLine[gof_summaryLineCounter++]=" ";


}
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void FinishGOFsummaryReport(double customMaxCriterion) {
   gof_summaryLine[gof_summaryLineCounter++]=" ";
   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("%.2f    | %s",customMaxCriterion==EMPTY_VALUE?-999:customMaxCriterion,"Custom Max Obj Value");
   gof_summaryLine[gof_summaryLineCounter++]=StringFormat("*** Sim elapsed time = %.1f sec",GetMicrosecondCount()/1e6);

// display all comments on the chart whe running the visual mode.
// a print out will also appear in the Journal tab in the visual window

   int Dx = 6;
   int Dy = 6;
   int boxWidth = 400;
   int lineHeight = 13;
   int nlines=gof_summaryLineCounter;
   int boxHeight = lineHeight*nlines+10;
   int fontSize = 8;

   int CHART_ID=0;
   int SUB_WINDOW=0;
   string name = "CommentBox";

   if(gof_drawSummary && ! ObjectCreate(CHART_ID,name,OBJ_RECTANGLE_LABEL,SUB_WINDOW,0,0)) {
      PrintFormat("===:%s():Falied to create OBJ_RECTANGLE_LABEL. Fake Error code: [%d]",__FUNCTION__,-1);
   }
   ObjectSetInteger(CHART_ID, name, OBJPROP_CORNER, CORNER_LEFT_LOWER);
   ObjectSetInteger(CHART_ID, name, OBJPROP_XDISTANCE, Dx);
   ObjectSetInteger(CHART_ID, name, OBJPROP_YDISTANCE, boxHeight + Dy);
   ObjectSetInteger(CHART_ID, name, OBJPROP_XSIZE, boxWidth);
   ObjectSetInteger(CHART_ID, name, OBJPROP_YSIZE, boxHeight);
   ObjectSetInteger(CHART_ID, name, OBJPROP_BGCOLOR, clrBlack);
   ObjectSetInteger(CHART_ID, name, OBJPROP_BORDER_TYPE, OBJPROP_BORDER_COLOR);
   ObjectSetInteger(CHART_ID, name, OBJPROP_BORDER_COLOR, clrWhite);

   if(gof_printSummary)Print("================= OnTester GOF Summary ===================");

   for(int i=0; i<nlines; i++) {
      if(gof_drawSummary) {
         string objName="name_"+IntegerToString(i);
         if(!ObjectCreate(CHART_ID, objName, OBJ_LABEL, 0, 0, 0))
            return;
         ObjectSetInteger(CHART_ID, objName, OBJPROP_ANCHOR, ANCHOR_LEFT_UPPER);
         ObjectSetInteger(CHART_ID, objName, OBJPROP_CORNER, CORNER_LEFT_LOWER);
         ObjectSetInteger(CHART_ID, objName, OBJPROP_XDISTANCE, Dx + 6);
         ObjectSetInteger(CHART_ID, objName, OBJPROP_YDISTANCE, Dy - 5 + boxHeight - i*lineHeight);
         ObjectSetInteger(CHART_ID, objName, OBJPROP_COLOR, clrWhite);
         ObjectSetInteger(CHART_ID, objName, OBJPROP_FONTSIZE, fontSize);
         ObjectSetString(CHART_ID, objName, OBJPROP_TEXT, gof_summaryLine[i]);
      }
      if(gof_printSummary)Print(gof_summaryLine[i]);
   }
   if(gof_printSummary)Print("=====================================================");


   if(gof_drawSummary) {
      for(int i=0; i<nlines; i++) {
         name = "Value_" +IntegerToString(i);
         if(!ObjectCreate(CHART_ID, name, OBJ_LABEL, 0, 0, 0))
            return;
         ObjectSetInteger(CHART_ID, name, OBJPROP_ANCHOR, ANCHOR_RIGHT_UPPER);
         ObjectSetInteger(CHART_ID, name, OBJPROP_CORNER, CORNER_LEFT_LOWER);
         ObjectSetInteger(CHART_ID, name, OBJPROP_XDISTANCE, Dx + boxWidth - 5);
         ObjectSetInteger(CHART_ID, name, OBJPROP_YDISTANCE, Dy - 5 + boxHeight - i*lineHeight);
         ObjectSetInteger(CHART_ID, name, OBJPROP_COLOR, clrDarkRed);
         ObjectSetInteger(CHART_ID, name, OBJPROP_FONTSIZE, fontSize);
         ObjectSetString(CHART_ID, name, OBJPROP_TEXT, " ");
      }
   }
   return;
}

//+------------------------------------------------------------------+
